//=============================================================================
// Jetpack.
// Revised by AnthraX & Scarface (Shiva)
// HIGOR: Client states and clientside movement correction
// HIGOR: Compatible with Extended Client Move
//=============================================================================
class Jetpack extends sgSuit
    config;

#exec OBJ LOAD File=Extro.uax

/*--- Data. -----------------------------------------------------------------*/

var() float         MaxFuel,
                    RechargeRate,
                    Thrust,
                    MaxVel,
                    HorizThrust,
                    MaxHorizVel,
                    ThrustFrequency;
var float           Fuel;           // Fuel left, measured in seconds
var Effects         Trail;
var bool            SimFall;
var float           ThrustTimer;
var float           OldLTS;	   // TR independent jetpack

var config bool     bNoAutoBind;

var float aTimer;
var float deltaUpdate; //Fuel variation since last update
var float LastFuel; //Client last fuel

var JetPushPool PushPool;
var bool bDelayActivate;
var bool bOldFalling;
var bool bLogNow;

replication
{
    reliable if ( Role < ROLE_Authority )
        ServerSetJetpack, ServerToggleJetpack;

    reliable if ( Role == ROLE_Authority )
        ClientSetBind, ClientDeactivate;

    unreliable if ( Role == ROLE_Authority )
        Trail, Fuel, SimFall, MaxFuel, RechargeRate, Thrust, MaxVel,
        HorizThrust, MaxHorizVel, ThrustFrequency;
}

function PostBeginPlay()
{
    Super.PostBeginPlay();
    bAutoActivate = false;
    Fuel = MaxFuel;
}

simulated function PostNetBeginState()
{
	GotoState('ClientDeActivated');
}

function GiveTo(Pawn other)
{
    Super.GiveTo(other);
    if ( Owner == other )
    {
        if ( PlayerPawn(Owner) != None )
        {
            ClientSetBind();
            ClientDeactivate();
        }
        GotoState('DeActivated');
    }
}

/*--- Client Functions. -----------------------------------------------------*/
function AnnounceAll(string sMessage)
{
    local Pawn p;

    for ( p = Level.PawnList; p != None; p = p.nextPawn )
	    if ( (p.bIsPlayer || p.IsA('MessagingSpectator')) &&
          p.PlayerReplicationInfo != None  )
		    p.ClientMessage(sMessage);
}

simulated function ClientSetBind()
{
    local int key;
    local string keyName, bind, bindCaps;
    local PlayerPawn playerOwner;

    playerOwner = PlayerPawn(Owner);
    if ( playerOwner == None || bNoAutoBind )
        return;

    for ( key = 0; key < 255; key++ )
    {
        keyName = playerOwner.ConsoleCommand("KEYNAME"@key);
        bind = playerOwner.ConsoleCommand("KEYBINDING"@keyName);
        bindCaps = Caps(bind);

        if ( Left(bindCaps, 4) == "JUMP" || InStr(bindCaps, " JUMP") != -1 ||
          InStr(bindCaps, "|JUMP") != -1 )
        {
            if ( Left(bindCaps, 10) != "SETJETPACK" &&
              InStr(bindCaps, " SETJETPACK") == -1 &&
              InStr(bindCaps, "|SETJETPACK") == -1 &&
              Left(bindCaps, 13) != "TOGGLEJETPACK" &&
              InStr(bindCaps, " TOGGLEJETPACK") == -1 &&
              InStr(bindCaps, "|TOGGLEJETPACK") == -1 )
            {
                bind = "SetJetpack 1|"$bind$"|OnRelease SetJetpack 0";
                playerOwner.ConsoleCommand("SET INPUT"@keyName@bind);
            }
        }
    }

    SaveConfig();
}

simulated function ClientDeactivate()
{
    if ( Role == ROLE_Authority )
        return;

    GotoState('ClientDeActivated');
//    Log("Jetpack forcing deactivation");
}

/*--- Console Functions. ----------------------------------------------------*/

//Client simulation
exec simulated function ToggleJetpack()
{
	if ( Level.NetMode == NM_Client )
	{
		if ( IsInState('ClientActivated') )
			GotoState('ClientDeActivated');
		else
			bDelayActivate = true;
	}
	ServerToggleJetpack();
}
exec simulated function SetJetpack(bool enabled)
{
	if ( enabled && (Level.NetMode == NM_Client) )
	{
		bDelayActivate = true;
		ServerSetJetpack( true);
		return;
	}

	if ( Owner.Physics != PHYS_Falling )
		return;

	if ( Level.NetMode == NM_Client )
	{
		if ( !enabled )
			GotoState('ClientDeActivated');
	}
	ServerSetJetpack( enabled);
}


function ServerToggleJetpack()
{
	if ( IsInState('Activated') )
		GotoState( 'DeActivated' );
	else if ( Owner.Physics == PHYS_Falling )
		GotoState( 'Activated' );
}

function ServerSetJetpack(bool enabled)
{
	if ( enabled )
	{
		if ( Owner.Physics == PHYS_Falling && !IsInState('Activated') )
			GotoState('Activated');
	}
	else if ( !IsInState('DeActivated') )
		GotoState('DeActivated');
}

simulated event Tick( float DeltaTime)
{
	if ( bDelayActivate )
	{
		if ( bOldFalling && Owner.Physics == PHYS_Falling )
			GotoState('ClientActivated');
		bDelayActivate = false;
	}
	if ( (Owner != none) && (Owner.Role == ROLE_AutonomousProxy) )
		bOldFalling = (Owner.Physics == PHYS_Falling);
}

/*-----------------------------------------------------------------------------
 * STATE Activated.
 * --------------------------------------------------------------------------*/

state Activated
{
	function BeginState()
	{
		local UT_Invisibility inv;
		if ( Role == ROLE_Authority )
		{
			if ( Owner.Physics != PHYS_Falling || Fuel <= 0 )
			{
				GotoState('DeActivated');
				return;
			}

			bActive = true;

			// Effects
			inv = UT_Invisibility(Pawn(Owner).FindInventoryType(class'UT_Invisibility'));
			if ( Trail == None )
				if ( inv == None || inv.charge < 160)
					Trail = Spawn(class'JetTrail', Owner);
		}

		// Physics
		if ( PlayerPawn(Owner) != None && SimFall )
			Pawn(Owner).bSimFall = true;
		
		aTimer = 1 / ThrustFrequency;
	}


	function EndState()
	{
		local sgSpeed sp;

		sp = sgSpeed(Pawn(Owner).FindInventoryType(class'sgSpeed'));
		if (sp != None)
			sp.GotoState('Activated');

		if ( Role == ROLE_Authority )
		{
			bActive = false;
			// Effects
			if ( Trail != None )
			{
				Trail.Destroy();
				Trail = None;
			}
		}

		// Physics
		if ( PlayerPawn(Owner) != None )
			Pawn(Owner).bSimFall = false;
	}
	
	event Tick( float Delta)
	{
		Global.Tick( Delta);
		if ( aTimer <= 0 )
		{
			aTimer += 1 / ThrustFrequency;
			UserTimer();
		}
		aTimer -= Delta;
		Fuel = FMax(Fuel - Delta, 0);
	}
	
	function UserTimer()
	{
		local vector horizVel;
		local rotator rot;
		local sgSpeed sp;
		
		if ( Role == ROLE_Authority && (Owner.Physics != PHYS_Falling || Fuel <= 0) )
		{
			GotoState('DeActivated');
			return;
		}

		if ( bLogNow )
			Log("JetLog: OldVel="$Owner.Velocity);

		sp = sgSpeed(Pawn(Owner).FindInventoryType(class'sgSpeed'));
		if (sp != None)
			sp.GotoState('DeActivated');
	
		horizVel = Owner.Velocity * vect(1,1,0);
		rot.Yaw = Owner.Rotation.Yaw;
		if ( cos((rotator(horizVel).Yaw - rot.Yaw)/32768*pi) *
		  VSize(horizVel) < MaxHorizVel )
			Owner.Velocity += vector(rot) * HorizThrust * 1 / ThrustFrequency;

		if ( Owner.Velocity.Z < MaxVel )
			Owner.Velocity.Z = FMin(Owner.Velocity.Z + Thrust * 1 / ThrustFrequency, MaxVel);

		if ( bLogNow )
			Log("JetLog: NewVel="$Owner.Velocity);
	}
}

simulated state ClientActivated
{
	simulated function BeginState()
	{
		aTimer = 1 / ThrustFrequency;
		DeltaUpdate = 0;
		CheckPool();
	}


	simulated function EndState()
	{
		local sgSpeed sp;

		sp = sgSpeed(Pawn(Owner).FindInventoryType(class'sgSpeed'));
		if (sp != None)
			sp.GotoState('Activated');

	}
	
	simulated event Tick( float Delta)
	{
		if ( aTimer <= 0 )
		{
			aTimer += 1 / ThrustFrequency;
			UserTimer();
		}
		aTimer -= Delta;

		if ( LastFuel != Fuel ) //Server update fuel!
		{
			Fuel += deltaUpdate;
			deltaUpdate = 0;
		}
		deltaUpdate -= Delta;
		Fuel = FMax(Fuel - Delta, 0);
		LastFuel = Fuel;
		Global.Tick( Delta);
	}
	
	simulated function UserTimer()
	{
		local vector horizVel;
		local rotator rot;
		local sgSpeed sp;
		local vector VelAdd;
		
		if ( Fuel <= 0 )
		{
			GotoState('ClientDeActivated');
			return;
		}

		if ( bLogNow )
			Log("JetLog: OldVel="$Owner.Velocity);

		sp = sgSpeed(Pawn(Owner).FindInventoryType(class'sgSpeed'));
		if (sp != None)
			sp.GotoState('DeActivated');
	
		horizVel = Owner.Velocity * vect(1,1,0);
		rot.Yaw = Owner.Rotation.Yaw;

		VelAdd = Owner.Velocity;
		if ( cos((rotator(horizVel).Yaw - rot.Yaw)/32768*pi) * VSize(horizVel) < MaxHorizVel )
			VelAdd += vector(rot) * HorizThrust * 1 / ThrustFrequency;
		if ( VelAdd.Z < MaxVel )
			VelAdd.Z = FMin(VelAdd.Z + Thrust * 1 / ThrustFrequency, MaxVel);
		PushPool.AddNewPush( VelAdd - Owner.Velocity);
		Owner.Velocity = VelAdd;

		if ( bLogNow )
			Log("JetLog: NewVel="$Owner.Velocity);
	}
}

/*-----------------------------------------------------------------------------
 * STATE DeActivated.
 * --------------------------------------------------------------------------*/

state DeActivated
{
	event BeginState()
	{
		ClientDeactivate();
	}
	simulated event Tick( float DeltaTime)
	{
		Fuel = FMin(Fuel + RechargeRate * deltaTime, MaxFuel);
	}
Begin:
}

simulated state ClientDeActivated
{
	simulated event Tick( float DeltaTime)
	{
		Fuel = FMin(Fuel + RechargeRate * deltaTime, MaxFuel);
		Global.Tick( DeltaTime);
	}
Begin:
}



//Clientside corrections
simulated function CheckPool()
{
	if ( PlayerPawn(Owner) == none || ViewPort(PlayerPawn(Owner).Player) == none || PushPool != none )
		return;
		
	ForEach AllActors (class'JetPushPool', PushPool)
		return;
	PushPool = Spawn(class'JetPushPool', Owner);
}


defaultproperties
{
    MaxFuel=3.00
    RechargeRate=0.80
    Thrust=1500.00
    MaxVel=350.00
    HorizThrust=250.00
    MaxHorizVel=400.00
    ThrustFrequency=20.00
    bActivatable=True
    bDisplayableInv=True
    PickupMessage="You got the Jetpack."
    ItemName="Jetpack"
    RespawnTime=60.00
    PickupViewMesh=LodMesh'UnrealI.AsbSuit'
    ProtectionType1=ProtectNone
    ProtectionType2=ProtectNone
    Charge=150
    MaxDesireability=3.00
    PickupSound=Sound'UnrealI.Pickups.FieldSnd'
    Icon=Texture'UnrealShare.Icons.I_ShieldBelt'
    bOwnerNoSee=True
    bTrailerSameRotation=True
    Mesh=LodMesh'Botpack.ShieldBeltMeshM'
    bGameRelevant=True
    CollisionRadius=25.00
    CollisionHeight=10.00
}
