//=============================================================================
// Mine 
// Written by nOs*Badger
// Revised for Monster Madness by nOs*Wildcard
// Revised by SK
// Optimized and de-exploited by Higor
//=============================================================================
class Mine extends sgBuilding;

var Pawn CheckOn[8];
var int iCheckOn;
var sgMineTrigger MyTrigger;
var float EffectGlow;

replication
{
	reliable if ( bNetInitial && (Role == ROLE_Authority) )
		EffectGlow;
}

//First event in creation order?
event Spawned()
{
	local PlayerStart P;
	local byte aTeam;

	Super.Spawned();

	if ( (Pawn(Owner) == none) || (Pawn(Owner).PlayerReplicationInfo == none) )
		return;

	aTeam = Pawn(Owner).PlayerReplicationInfo.Team;
	ForEach RadiusActors (class'PlayerStart', P, 80)
	{
		if ( P.TeamNumber != aTeam )
		{
			Destroy();
			return;
		}
	}
}

function CheckThis( Pawn Other)
{
	if ( iCheckOn >= 8 )
		return; //8 pawns on mine!!??

	CheckOn[iCheckOn++] = Other;

	if ( FastTrace(Other.Location) )
		Timer();
}

function bool ShouldDetonate()
{
	local int i;
	local bool bSuccess;

	For ( i=iCheckOn-1 ; i>=0 ; i-- )
	{
		if ( !MyTrigger.IsTouching( CheckOn[i]) ) //Clean up the whole list before returning
		{
			CheckOn[i] = CheckOn[--iCheckOn];
			CheckOn[iCheckOn] = none;
			continue;
		}

		if ( !bSuccess && (CheckOn[i].Health > 0) && (VSize( CheckOn[i].Location - Location) < (75 + CheckOn[i].CollisionRadius)) && FastTrace(CheckOn[i].Location) )
			bSuccess = true;
	}
	return bSuccess;
}

function TriggerSize()
{
	MyTrigger.SetCollisionSize( 75, 75);
}

event Destroyed()
{
	if ( MyTrigger != none )
		MyTrigger.Destroy();
	Super.Destroyed();
}

function PostBuild()
{
	Super.PostBuild();
	Texture = Texture'SKFlare';
}


simulated function string KillMessage( name damageType, pawn Other )
{
    local sgWarhead sgNuke;
    local string s;
    local string sWarhead;
    local pawn p;

    p = GetAwardedPlayer();

    s=p.PlayerReplicationInfo.PlayerName;

    if (s!="" && s!=" ")
         s=" built by"@s;

    if (sgNukeLauncher(Other.FindInventoryType(class'sgNukeLauncher')).AmmoType.AmmoAmount == 1)
	{
        sWarhead = ". "$Other.PlayerReplicationInfo.PlayerName@"was carrying a WARHEAD!!!";
	sgPRI(p.PlayerReplicationInfo).AddRU(500);
	sgPRI(p.PlayerReplicationInfo).sgInfoWarheadKiller++;
	}
	
	else if (sgNukeLauncher(Other.FindInventoryType(class'sgNukeLauncher')).AmmoType.AmmoAmount > 1)
	{
        sWarhead = ". "$Other.PlayerReplicationInfo.PlayerName@"was carrying TWO WARHEADS!!!";
	sgPRI(p.PlayerReplicationInfo).AddRU(1000);
	sgPRI(p.PlayerReplicationInfo).sgInfoWarheadKiller++;
	sgPRI(p.PlayerReplicationInfo).sgInfoWarheadKiller++;
	}

    else
        sWarhead = ".";
    return (""@Other.PlayerReplicationInfo.PlayerName@"was killed by a"@BuildingName$s$sWarhead);
}

function bool ShouldAttack(Pawn enemy)
{
	if ( ScriptedPawn(enemy) != None )
		return true;
	if ( sgBuilding(enemy) != None )
	{
		if ( sgBuilding(enemy).Team == Team || sgBuilding(enemy).Energy < 0 )
			return false;
	}
	else if ( enemy.PlayerReplicationInfo == None ||
			enemy.PlayerReplicationInfo.Team == Team ||
			!enemy.bProjTarget )
		return false;
	return true;
}


function Damage()
{
	local int i, j, iP;
	local Pawn p, pAward, pList[16];

	PlaySound(Sound'SharpExplosion',, 4.0);
	pAward = GetAwardedPlayer();

	//Hack fix, This iterator doesn't include CollisionRadius, use 25 extra for max detection (for monster support)!!
	ForEach VisibleCollidingActors (class'Pawn', p, 75+(int(Grade)*25)+25,,true) 
	{
		if ( (p.Health > 0) && ShouldAttack(P) )
		{
			pList[iP++] = p;
			if ( iP >= arraycount(pList) )
				break;
		}
	}

	i=Grade;	
	While ( i-- >= 0 )
		Spawn (class'MineExplosion',,, Location + VRand() * 40);

	While ( j < iP )
	{
		i = 0;
		While ( VSize( Location - pList[j].Location ) > 75+17+(i*25) )
			i++;
		if ( i > Grade )
		{
			j++;
			continue;
		}
		i = 1 + int(Grade) - i;

		pList[j].TakeDamage( (Grade+1) * i * 15, instigator, normal( Location - pList[j].Location) * 0.5 * (pList[j].CollisionHeight + pList[j].CollisionRadius), vect(0,0,0), 'mine');
		sgPRI(pAward.PlayerReplicationInfo).AddRU( (Grade+1) * i * 4);
		j++;
	}
	sgPRI(pAward.PlayerReplicationInfo).AddRU(50);

	Destroy();
}



function AnnounceAll(string sMessage)
{
    local Pawn p;

    for ( p = Level.PawnList; p != None; p = p.nextPawn )
	    if ( (p.bIsPlayer || p.IsA('MessagingSpectator')) &&
          p.PlayerReplicationInfo != None  )
		    p.ClientMessage(sMessage);
}

function Pawn GetAwardedPlayer()
{
	return Pawn(Owner);
}

simulated function CompleteBuilding()
{
	local Pawn p;

	if ( Level.NetMode == NM_Client )
		Assert( Default.EffectGlow == 0.2);

	if ( Role != ROLE_Authority )
		return;

	if ( VSize( Location - MyTrigger.Location) > 10)
		MyTrigger.SetLocation( Location - vect(0,0,1) );
	if ( ShouldDetonate() && !bDisabledByEMP )
		Damage();
}

simulated function FinishBuilding()
{
    local int i;
    local sgMeshFX newFX;
    local vector spawnLocation;
	
	Texture=Texture'Botpack.FLAKAMMOLEDbase';

	if ( Role == ROLE_Authority )
	{
		Spawn(class'sgFlash');
		MyTrigger = Spawn(class'sgMineTrigger', none, 'sgminetrigger');
		MyTrigger.Master = self;
		TriggerSize();
		MyTrigger.SetLocation( Location - vect(0,0,1) );
	}

	if ( Level.NetMode == NM_DedicatedServer )
		return;

	spawnLocation = Location;
	spawnLocation.Z -= 10;
	if ( myFX == None && Model != None )
		for ( i = 0; i < numOfMFX; i++ )
		{
			newFX = Spawn(class'WildcardsMeshFX', Self,,,
				rotator(vect(0,0,0)));
			newFX.NextFX = myFX;
			myFX = newFX;
			myFX.Mesh = Model;
			myFX.DrawScale = DSofMFX;
			myFX.RotationRate.Pitch = MFXrotX.Pitch*FRand();
			myFX.RotationRate.Roll = MFXrotX.Roll*FRand();
			myFX.RotationRate.Yaw = MFXrotX.Yaw*FRand();
			myFX.AmbientGlow=1;
			myFX.Style = STY_Translucent;
			myFX.ScaleGlow = EffectGlow;
			//myFX.LODBias=0.200000;
		}
}

defaultproperties
{
     bOnlyOwnerRemove=True
     BurnPerSecond=5
     EffectGlow=0.2
     BuildingName="Mine"
     BuildCost=125
     UpgradeCost=15
     BuildTime=1.000000
     MaxEnergy=2000.000000
     SpriteScale=0.250000
     Model=LodMesh'Botpack.DiscStud'
     SpriteRedTeam=Texture'Botpack.FLAKAMMOLEDbase'
     SpriteBlueTeam=Texture'Botpack.FLAKAMMOLEDbase'
     SpriteGreenTeam=Texture'Botpack.FLAKAMMOLEDbase'
     SpriteYellowTeam=Texture'Botpack.FLAKAMMOLEDbase'
     MFXrotX=(Yaw=10000)
     AmbientGlow=0
     CollisionHeight=10.000000
     LightBrightness=1
     Visibility=15
}
